@@ -9,11 +9,8 @@
 // <ranges>
 
 // UNSUPPORTED: c++03, c++11, c++14, c++17
-// UNSUPPORTED: libcpp-no-concepts
-// UNSUPPORTED: gcc-10
-// UNSUPPORTED: libcpp-has-no-incomplete-ranges
 
-#include <ranges>
+#include <concepts>
 
 #include "test_macros.h"
 
