@@ -7,7 +7,6 @@
 //===----------------------------------------------------------------------===//
 
 // UNSUPPORTED: c++03, c++11, c++14, c++17
-// UNSUPPORTED: libcpp-no-concepts
 // UNSUPPORTED: libcpp-has-no-incomplete-ranges
 
 // join_view() requires default_initializable<V> = default;
@@ -15,13 +14,26 @@
 #include <cassert>
 #include <ranges>
 
-#include "test_macros.h"
 #include "types.h"
 
+struct DefaultView : std::ranges::view_base {
+  int i; // deliberately uninitialised
+
+  ChildView* begin() const;
+  ChildView* end() const;
+};
 
 constexpr bool test() {
-  std::ranges::join_view<ParentView<ChildView>> jv;
-  assert(std::move(jv).base().ptr_ == globalChildren);
+  {
+    std::ranges::join_view<ParentView<ChildView>> jv;
+    assert(std::move(jv).base().ptr_ == globalChildren);
+  }
+
+  // Default constructor should value initialise underlying view
+  {
+    std::ranges::join_view<DefaultView> jv;
+    assert(jv.base().i == 0);
+  }
 
   static_assert( std::default_initializable<std::ranges::join_view<ParentView<ChildView>>>);
   static_assert(!std::default_initializable<std::ranges::join_view<CopyableParent>>);
